/* Author: Tobi Vollebregt */

/*  Marsworm -- A small game written for the TINS Is Not SpeedHack 2003
 *  Copyright (C) 2003  Tobi Vollebregt
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  See `License.txt', which contains a verbatim copy of the
 *  GNU General Public License, for details.
 *
 *  Please send your reaction to: tobivollebregt@hotmail.com
 */

#include <allegro.h>
#include <string.h>
#include <time.h>
#include "moon.h"

#define TICKS_PER_SECOND 50
#define SECONDS_PER_TICK (1.0 / TICKS_PER_SECOND)

static BITMAP* buffer;
static volatile int speed_counter, fps, frame;
static int subdiv=3;
int worm_type=0;  // 0 = radioactive, 1 = jelly like substance
static int topscore=0;

// also used in options menu
static int card = GFX_AUTODETECT;
static int width = 0, height = 0;
static int color_depth;

static void ticker()
{
    ++speed_counter;
 
}
END_OF_STATIC_FUNCTION(ticker);

static void fps_proc()
{
    fps = frame;
    frame = 0;
}
END_OF_STATIC_FUNCTION(fps_proc);

void exitmsg(const char* fmt, ...)
{
    char buf[4096];
    va_list argp;

    va_start(argp, fmt);
    uvszprintf(buf, sizeof(buf), fmt, argp);
    va_end(argp);

    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    allegro_message("%s", buf);

    exit(1);
}

void initialize(int, const char* const*)
{
    bool installed_tmk = false;

    if (width == 0 || height == 0)
    {
        if (set_gfx_mode(GFX_SAFE, 320, 200, 0, 0) < 0)
            exitmsg("Error setting 320x200 graphics mode:\n%s", allegro_error);

        installed_tmk = true;
        if (install_timer() < 0)
            exitmsg("Couldn't install timer.");

        if (install_mouse() < 0)
            exitmsg("Couldn't install mouse.");

        if (install_keyboard() < 0)
            exitmsg("Couldn't install keyboard.");

        PALETTE pal;
        int i;

        for (i=0; i<256; ++i) pal[i].r = pal[i].g = pal[i].b = (255-i) >> 2;
        set_palette(pal);
        set_mouse_sprite(0);

        for (i=0; i<256; ++i) pal[i].r = pal[i].g = pal[i].b = i >> 2;
        set_palette(pal);

        gui_fg_color = 0;
        gui_bg_color = 200;
        gui_mg_color = 127;

        width = 640;
        height = 480;
        color_depth = desktop_color_depth();

        if (gfx_mode_select_ex(&card, &width, &height, &color_depth) == 0)
            exit(0);

        if (color_depth == 8)
            exitmsg("256 color (8 bpp) video mode not supported.");
    }

    set_color_depth(color_depth);
    if (set_gfx_mode(card, width, height, 0, 0) < 0)
        exitmsg("Error setting %dx%dx%d graphics mode:\n%s\n"
        "Try delete the file `marsworm.opt' in the program's directory.",
            width, height, color_depth, allegro_error);

    if (!installed_tmk)
    {
        if (install_timer() < 0)
            exitmsg("Couldn't install timer.");

        if (install_mouse() < 0)
            exitmsg("Couldn't install mouse.");

        if (install_keyboard() < 0)
            exitmsg("Couldn't install keyboard.");
    }

    gui_fg_color = 0;
    gui_bg_color = makecol(200, 200, 200);
    //gui_mg_color = ;

    clear_bitmap(screen);
    text_mode(0);
    textout_centre(screen, font, "INITIALIZING...", SCREEN_W / 2,
            (SCREEN_H - text_height(font)) / 2, makecol(255, 255, 255));

    if (!(buffer = create_bitmap(SCREEN_W, SCREEN_H)))
        exitmsg("Not enough memory for double buffering.");

    LOCK_VARIABLE(speed_counter);
    LOCK_FUNCTION(ticker);
    LOCK_VARIABLE(fps);
    LOCK_VARIABLE(frame);
    LOCK_FUNCTION(fps_proc);

    if (install_int_ex(ticker, BPS_TO_TIMER(TICKS_PER_SECOND)) < 0 ||
        install_int_ex(fps_proc, BPS_TO_TIMER(1)) < 0)
        exitmsg("Not enough room to install timer interrupts.");
}

bool do_menu(bool init=false)
{
    /////// this gui is made the last 4 hours

    const char* worm_type_text[] = {
        "&Snake: Radioactive waste",
        "&Snake: Jelly like substance"
    };
    const char* lod_text[] = {
        "&Level of Detail: Low",
        "&Level of Detail: Normal",
        "&Level of Detail: High"
    };

    DIALOG dlg[] = {
    /* int (*proc)(), int x, y, w, h, fg, bg, key, flags, d1, d2, void *dp, *dp2, *dp3 */
    { d_button_proc, 0, 0, 240, 20, 0, 0, 'p', D_EXIT, 0, 0, (void*) "&Play", 0, 0 },
    { d_button_proc, 0, 30, 240, 20, 0, 0, 'o', D_EXIT, 0, 0, (void*) "&Options", 0, 0 },
    { d_button_proc, 0, 60, 240, 20, 0, 0, 'q', D_EXIT, 0, 0, (void*) "&Quit", 0, 0 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }};

    DIALOG opt_dlg[] = {
    /* int (*proc)(), int x, y, w, h, fg, bg, key, flags, d1, d2, void *dp, *dp2, *dp3 */
    { d_button_proc, 0, 0, 240, 20, 0, 0, 'v', D_EXIT, 0, 0, (void*) "&Video modes", 0, 0 },
    { d_button_proc, 0, 30, 240, 20, 0, 0, 's', D_EXIT, 0, 0, 0, 0, 0 },
    { d_button_proc, 0, 60, 240, 20, 0, 0, 'l', D_EXIT, 0, 0, 0, 0, 0 },
    { d_button_proc, 0, 90, 240, 20, 0, 0, 'b', D_EXIT, 0, 0, (void*) "&Back", 0, 0 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }};

    BITMAP* bg = create_bitmap(SCREEN_W, SCREEN_H);
    if (!bg) exitmsg("not enough memory");
    clear_bitmap(bg);

    if (init)
    {
        BITMAP* mars = init ? load_bitmap("mars.bmp", 0) : 0;
        if (!mars) exitmsg("%s: %s", "mars.bmp", ustrerror(errno));

        stretch_blit(mars, bg, 0, 0, mars->w, mars->h, (bg->w-2*mars->w)/2, (bg->h-2*mars->h)/2, 2*mars->w, 2*mars->h);
        destroy_bitmap(mars);
        blit(bg, screen, 0, 0, 0, 0, bg->w, bg->h);
    }
    else blit(screen, bg, 0, 0, 0, 0, bg->w, bg->h);

    centre_dialog(dlg);
    centre_dialog(opt_dlg);
    set_dialog_color(dlg, gui_fg_color, gui_bg_color);
    set_dialog_color(opt_dlg, gui_fg_color, gui_bg_color);

    for (;;)
    {
        int c = do_dialog(dlg, 0);

        if (bg->w != SCREEN_W || bg->h != SCREEN_H)
            stretch_blit(bg, screen, 0, 0, bg->w, bg->h, 0, 0, SCREEN_W, SCREEN_H);
        else blit(bg, screen, 0, 0, 0, 0, bg->w, bg->h);

        if (c == 0) return true;
        if (c == -1 || c == 2) return false;

        if (c == 1)
        {
            for (;;)
            {
                opt_dlg[1].dp = (void*) worm_type_text[worm_type];
                opt_dlg[2].dp = (void*) lod_text[subdiv-2];

                c = do_dialog(opt_dlg, 0);

                if (bg->w != SCREEN_W || bg->h != SCREEN_H)
                    stretch_blit(bg, screen, 0, 0, bg->w, bg->h, 0, 0, SCREEN_W, SCREEN_H);
                else blit(bg, screen, 0, 0, 0, 0, bg->w, bg->h);

                if (c == 3 || c == -1) break;
                if (c == 1) worm_type = 1-worm_type;
                if (c == 2) if (++subdiv == 5) subdiv = 2;

                if (c == 0)
                {
                    int card0 = card, width0 = width, height0 = height, color_depth0 = color_depth;

                    if (gfx_mode_select_ex(&card0, &width0, &height0, &color_depth0))
                    {
                        if (color_depth == 8)
                        {
                            alert("256 color (8 bpp) video mode not supported.", 0, 0, "&Oops", 0, 'o', 0);
                            continue;
                        }

                        set_color_depth(color_depth0);
                        if (set_gfx_mode(card0, width0, height0, 0, 0) < 0)
                        {
                            char buf[512], buf2[512];
                            uszprintf(buf, sizeof(buf), "Error setting %dx%dx%d graphics mode:", width0, height0, color_depth0);
                            ustrzcpy(buf2, sizeof(buf2), allegro_error);

                            set_color_depth(color_depth);
                            if (set_gfx_mode(card, width, height, 0, 0) < 0)
                                exitmsg("Error setting %dx%dx%d graphics mode:\n%s",
                                        width, height, color_depth, allegro_error);

                            gui_fg_color = 0;
                            gui_bg_color = makecol(200, 200, 200);

                            if (bg->w != SCREEN_W || bg->h != SCREEN_H)
                                stretch_blit(bg, screen, 0, 0, bg->w, bg->h, 0, 0, SCREEN_W, SCREEN_H);
                            else blit(bg, screen, 0, 0, 0, 0, bg->w, bg->h);

                            alert(buf, buf2, 0, "&Oops", 0, 'o', 0);
                        }
                        else
                        {
                            if (color_depth != color_depth0)
                            {
                                BITMAP* new_bg = create_bitmap(bg->w, bg->h);

                                if (!new_bg) exitmsg("not enough memory");
                                blit(bg, new_bg, 0, 0, 0, 0, bg->w, bg->h);
                                destroy_bitmap(bg);
                                bg = new_bg;
                            }

                            if (bg->w != SCREEN_W || bg->h != SCREEN_H)
                                stretch_blit(bg, screen, 0, 0, bg->w, bg->h, 0, 0, SCREEN_W, SCREEN_H);
                            else blit(bg, screen, 0, 0, 0, 0, bg->w, bg->h);

                            card = card0;
                            width = width0;
                            height = height0;
                            color_depth = color_depth0;
                            destroy_bitmap(buffer);
                            buffer = create_bitmap(width, height);
                            if (!buffer) exitmsg("Not enough memory for double buffering.");

                            gui_fg_color = 0;
                            gui_bg_color = makecol(200, 200, 200);
                            centre_dialog(dlg);
                            centre_dialog(opt_dlg);
                            set_dialog_color(dlg, gui_fg_color, gui_bg_color);
                            set_dialog_color(opt_dlg, gui_fg_color, gui_bg_color);
                        }
                    }
                }
            }
        }
    }
}

int main(int argc, const char* const* argv)
{
    srand((unsigned)time(0));
    allegro_init();

    PACKFILE* fp = pack_fopen("marsworm.opt", "r");
    if (fp)
    {
        subdiv = pack_mgetl(fp);
        worm_type = pack_mgetl(fp);
        topscore = pack_mgetl(fp);
        card = pack_mgetl(fp);
        width = pack_mgetl(fp);
        height = pack_mgetl(fp);
        color_depth = pack_mgetl(fp);
        pack_fclose(fp);
    }

    initialize(argc, argv);

    Moon* moon=0;
    bool pressed = false;

    if (!do_menu(true)) goto end;

    moon = new Moon;
    moon->initialize(subdiv);

    fps = frame = speed_counter = 0;

    for (;;)
    {
        ++frame;

        while (speed_counter > 0)
        {
            speed_counter--;
            int c = moon->update();
            if (c != -1 || key[KEY_ESC])
            {
                while(key[KEY_ESC]) ;
                clear_keybuf();

                //char buf[256];
                text_mode(0);
                if (moon->getScore() <= topscore)
                    textprintf_centre(screen, font, SCREEN_W/2, 40, makecol(255, 255, 0),
                            "You failed MISERABLY. You got only %d points.", moon->getScore());
                else
                {
                    topscore = moon->getScore();
                    textprintf_centre(screen, font, SCREEN_W/2, 40, makecol(0, 255, 0),
                            "Hooray, you've got a TOPSCORE of %d points.", moon->getScore());
                }
                /*if (alert(buf, "Play another game?", 0, "&Yes", "&No", 'y', 'n') == 1)
                {
                }
                else goto end;*/
                if (!do_menu()) goto end;

                delete moon;
                moon = new Moon;
                moon->initialize(subdiv);
                fps = frame = speed_counter = 0;
            }
        }

        clear_bitmap(buffer);
        moon->draw(buffer);
        if (mouse_b & 1)
        {
            if (!pressed) { moon->click(mouse_x, mouse_y); pressed = true; }
        }
        else pressed = false;

        if (topscore) {text_mode(0);textprintf_centre(buffer, font, SCREEN_W/2, 10, makecol(255, 255, 255), "topscore: %d", topscore);}
        if (fps) {text_mode(0); textprintf(buffer, font, 0, 0, makecol(255, 255, 255), "%d fps", fps);}
        show_mouse(buffer);

        if (key[KEY_F12]) save_bitmap("screen.bmp", buffer, 0);

        blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
        show_mouse(screen);
    }
 end:
    if (moon) delete moon;

    fp = pack_fopen("marsworm.opt", "w");
    if (fp)
    {
        pack_mputl(subdiv, fp);
        pack_mputl(worm_type, fp);
        pack_mputl(topscore, fp);
        pack_mputl(card, fp);
        pack_mputl(width, fp);
        pack_mputl(height, fp);
        pack_mputl(color_depth, fp);
        pack_fclose(fp);
    }

    return 0;
}
END_OF_MAIN()
