/* Author: Tobi Vollebregt */

/*  Marsworm -- A small game written for the TINS Is Not SpeedHack 2003
 *  Copyright (C) 2003  Tobi Vollebregt
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  See `License.txt', which contains a verbatim copy of the
 *  GNU General Public License, for details.
 *
 *  Please send your reaction to: tobivollebregt@hotmail.com
 */

#ifndef __tjv_vector_h__
#define __tjv_vector_h__

#include <math.h>

#ifndef PI
 #define PI 3.14159265358979323846
#endif // !defined(PI)

#ifdef _MSC_VER
// float to double conversion error
#pragma warning(disable:4244) 
#endif

// operator|() implements crossproduct
// operator*() implements dotproduct

class Vector
{
 public:
    typedef float value_type;
    typedef value_type* pointer;
    typedef const value_type* const_pointer;

    static Vector Polar(value_type theta, value_type phi, value_type r=1.0);

    Vector() {}
    Vector(value_type n): x_(n), y_(n), z_(n) {}
    Vector(value_type x, value_type y, value_type z): x_(x), y_(y), z_(z) {}
    Vector(const_pointer p): x_(p[0]), y_(p[1]), z_(p[2]) {}

    Vector operator-() const { return Vector(-x_, -y_, -z_); }

    Vector operator+(const Vector& w) const { return Vector(x_+w.x(), y_+w.y(), z_+w.z()); }
    Vector operator-(const Vector& w) const { return Vector(x_-w.x(), y_-w.y(), z_-w.z()); }
    Vector operator|(const Vector& w) const { return Vector(y_*w.z()-z_*w.y(), z_*w.x()-x_*w.z(), x_*w.y()-y_*w.x()); }
    Vector operator*(value_type n) const { return Vector(x_*n, y_*n, z_*n); }
    inline friend Vector operator*(value_type n, const Vector& v) { return v*n; }
    value_type operator*(const Vector& w) const { return x_*w.x()+y_*w.y()+z_*w.z(); }

    Vector& operator=(value_type n) { x_=y_=z_=n; return *this; }
    Vector& operator=(const_pointer p) { x_=p[0]; y_=p[1]; z_=p[2]; return *this; }
    Vector& operator+=(const Vector &w) { x_+=w.x(); y_+=w.y(); z_+=w.z(); return *this; }
    Vector& operator-=(const Vector &w) { x_-=w.x(); y_-=w.y(); z_-=w.z(); return *this; }
    Vector& operator|=(const Vector &w) { return (*this = *this | w); }
    Vector& operator*=(value_type n) { x_*=n; y_*=n; z_*=n; return *this; }

    bool operator==(const Vector& w) const;
    bool operator!=(const Vector& w) const { return !operator==(w); }

    value_type magnitude() const { return sqrt(x_*x_+y_*y_+z_*z_); }
    value_type theta() const { value_type theta = atan2(y_, x_); return theta; }
    value_type phi() const { value_type phi = atan2(z_, sqrt(x_*x_+y_*y_)); return phi; }
    Vector& normalize() { return *this *= 1.0/magnitude(); }

    void clamp(value_type lo, value_type hi);

    void setxyz(value_type x, value_type y, value_type z) { x_=x; y_=y; z_=z; }
    void setPolar(value_type theta, value_type phi, value_type r=1.0);

    value_type cosx() const { return x_/magnitude(); }
    value_type cosy() const { return y_/magnitude(); }
    value_type cosz() const { return z_/magnitude(); }
    value_type x() const { return x_; }
    value_type y() const { return y_; }
    value_type z() const { return z_; }

    pointer adrx() { return &x_; }
    pointer adry() { return &y_; }
    pointer adrz() { return &z_; }
    const_pointer adrx() const { return &x_; }
    const_pointer adry() const { return &y_; }
    const_pointer adrz() const { return &z_; }

 private:
    value_type x_, y_, z_;
};

bool pointInPoly(const Vector& pt, const Vector* ptr, int size);

#endif // !defined(__tjv_vector_h__)
